/*
@license

dhtmlxGantt v.3.2.2 Stardard
This software is covered by GPL license. You also can obtain Commercial or Enterprise license to use it in non-GPL project - please contact sales@dhtmlx.com. Usage without proper license is prohibited.

(c) Dinamenta, UAB.
*/
gantt.config.auto_scheduling = false;
gantt.config.auto_scheduling_strict = false;

gantt._autoScheduleHandler = function(taskId, mode, task) {
    if (gantt.config.auto_scheduling && !this._autoscheduling_in_proggress) {
        var newTask = this.getTask(taskId);
        if(task.start_date.valueOf() != newTask.start_date.valueOf() ||
            ((task.end_date.valueOf() != newTask.end_date.valueOf() ||
            task.duration != newTask.duration)&& task.type != gantt.config.types.milestone)) {
            gantt.autoSchedule(newTask.id);
        }
    }
    return true;
};

gantt._linkChangeAutoScheduleHandler = function(linkId, link) {
    if (gantt.config.auto_scheduling && !this._autoscheduling_in_proggress) {
        gantt.autoSchedule(link.source);
    }
};

gantt._autoScheduleOnLoadHandler = function()
{
    if (gantt.config.auto_scheduling) {
        var strictFlag = gantt.config.auto_scheduling_strict;
        gantt.config.auto_scheduling_strict = false;
        gantt.autoSchedule();
        gantt.config.auto_scheduling_strict = strictFlag;
    }
};

gantt._checkCircularDependency = function (element, checkLink, usedElements) {
    if(usedElements[element.id]) {
        return false;
    }
    else {

        usedElements[element.id] = true;
    }
    var currentRefEl;

    for(var i=0; i<element.$source.length; i++) {
        if (element.$source[i] == checkLink.id) {continue;}
        var currentLink = this.getLink(element.$source[i]);
        currentRefEl = this.getTask(currentLink.target);
        if (!this._checkCircularDependency(currentRefEl, checkLink, usedElements)) {
            return false;
        }
    }
    if(element.id == checkLink.source)
    {
        currentRefEl = this.getTask(checkLink.target);
        if(!this._checkCircularDependency(currentRefEl,checkLink,usedElements)) {
            return false;
        }
    }

    var children = gantt.getChildren(element.id);
    for(var i = 0; i < children.length; i++){
        var refElement = this.getTask(children[i]);
        if(refElement.$target.length == 0) {
            if (!this._checkCircularDependency(refElement, checkLink, usedElements)) {
                return false;
            }
        }
    }
    return true;
};

gantt._checkParents = function(id, tasks){
    if(tasks[id]){
        return false;
    }

    if(id != gantt.config.root_id){
        var parent = gantt.getParent(id);
        if(parent != gantt.config.root_id) {
            return gantt._checkParents(parent, tasks);
        }
    }

    return true;
};

gantt._linkValidationHandler = function(id,link)
{
    var result = !(gantt.isChildOf(link.target, link.source) || gantt.isChildOf(link.source, link.target)) &&
        gantt._checkCircularDependency(gantt.getTask(link.source),link, {});

    if(!result) {
        gantt.callEvent("onLinkDependencyError", [link]);
    }

    return result;
};

gantt._lightBoxChangesHandler = function(taskId,task)
{

    if (gantt.config.auto_scheduling && !this._autoscheduling_in_proggress) {
        var oldTask = this.getTask(taskId);
        if(task.start_date.valueOf() != oldTask.start_date.valueOf() ||
            ((task.end_date.valueOf() != oldTask.end_date.valueOf() ||
            task.duration != oldTask.duration)&& task.type != gantt.config.types.milestone)) {
            gantt._autoschedule_lightbox_id = taskId;
        }
    }
    return true;
};

gantt._lightBoxSaveHandler = function(taskId, task){

    if (gantt.config.auto_scheduling && !this._autoscheduling_in_proggress) {
        if (gantt._autoschedule_lightbox_id && gantt._autoschedule_lightbox_id == taskId) {
            gantt._autoschedule_lightbox_id = null;
            gantt.autoSchedule(task.id);
        }
    }
    return true;
};

gantt.attachEvent("onAfterLinkUpdate", gantt._linkChangeAutoScheduleHandler);
gantt.attachEvent("onAfterLinkAdd", gantt._linkChangeAutoScheduleHandler);
gantt.attachEvent("onParse",gantt._autoScheduleOnLoadHandler);
gantt.attachEvent("onBeforeLinkAdd", gantt._linkValidationHandler);
gantt.attachEvent("onBeforeLinkUpdate", gantt._linkValidationHandler);
gantt.attachEvent("onBeforeTaskChanged", gantt._autoScheduleHandler);
gantt.attachEvent("onLightboxSave", gantt._lightBoxChangesHandler);
gantt.attachEvent("onAfterTaskUpdate",gantt._lightBoxSaveHandler);


gantt.autoSchedule = function(taskId) {
    if (gantt.callEvent("onBeforeAutoScheduling", [taskId]) !== false) {
        var upTasks = {};
        var task;
        if (taskId) {
            task = this.getTask(taskId);
            upTasks = gantt._autoScheduleByTask(task, upTasks);
        }
        else {
            var tasks = gantt.getChildren(gantt.config.root_id);
            for (var i = 0; i < tasks.length; i++) {
                task = this.getTask(tasks[i]);
                upTasks = gantt._autoScheduleByTask(task, upTasks);
            }
        }
        gantt._updateTasks(upTasks);
        gantt.callEvent("onAfterAutoScheduling", [taskId, upTasks]);
    }
};

gantt._autoScheduleByTask = function(task, updatedTasksList) {
    if (gantt.config.auto_scheduling_strict) {
        updatedTasksList = gantt._autoScheduleLinkedTasks(task, updatedTasksList, null, task.id);
    }
    else {
        updatedTasksList = gantt._autoScheduleLinkedTasks(task, updatedTasksList, null);
        if(task.parent && task.parent != gantt.config.root_id) {
            updatedTasksList = gantt._autoScheduleParents(this.getTask(task.parent), updatedTasksList);
        }
    }

    return updatedTasksList;
};

gantt._autoScheduleLinkedTasks = function(task, updatedTasks, currentParent, strictStart) {
    var i;

    if (task.auto_scheduling === undefined || task.auto_scheduling) {
        var taskMinPosition = gantt._getTaskMinimalStartDate(task);
        if (task.parent && task.parent != gantt.config.root_id) {
            var parentTask = this.getTask(task.parent);
            var projectMinPosition = gantt._getProjectMinimalStartDate(parentTask);
            if (taskMinPosition.date && projectMinPosition.date) {
                if (taskMinPosition.date < projectMinPosition.date) {
                    taskMinPosition = projectMinPosition;
                }
            }
            else {
                if (!taskMinPosition.date) {
                    taskMinPosition = projectMinPosition;
                }
            }
        }
        // taskMinPosition is an object which contains minimal date, link to object and source task.
        // E.g.
        // {date: Fri Apr 05 2013 00:00:00 GMT+0300, link: Object, source_task: Object}
        if (gantt.callEvent("onBeforeAutoSchedule", [task, taskMinPosition]) === false) {
            return updatedTasks;
        }

        if (gantt._isTask(task)) {
            if (taskMinPosition.date &&
                ((gantt.config.auto_scheduling_strict &&
                (strictStart != task.id)) || taskMinPosition.date > task.start_date)) {
                gantt._updateTaskPosition(task, taskMinPosition.date, task.duration);
                updatedTasks[task.id] = true;
            }
        }
        var taskChildren = this.getChildren(task.id);

        var isAnStrictParent = gantt._isElementAnStrictStartParent(strictStart,task.id);

        if (taskChildren.length && (!gantt.config.auto_scheduling_strict ||
            ((!isAnStrictParent) || (isAnStrictParent && taskMinPosition._mustMove)))) {
            for (i = 0; i < taskChildren.length; i++) {
                gantt._autoScheduleLinkedTasks(this.getTask(taskChildren[i]), updatedTasks, currentParent, strictStart);
            }
        }
        gantt.callEvent("onAutoSchedule", [task]);
    }

    for(i=0; i<task.$source.length; i++) {
        var currentLink = this.getLink(task.$source[i]);
        var referencedTask = this.getTask(currentLink.target);
        gantt._autoScheduleLinkedTasks(referencedTask, updatedTasks, currentParent,strictStart);
    }

    if (task.parent && currentParent && task.parent != gantt.config.root_id) {

        var taskParent = this.getTask(task.parent);
        if (gantt.config.auto_scheduling_strict) {
            gantt._autoScheduleParents(taskParent, updatedTasks, strictStart);
        }
        else {
            gantt._autoScheduleParents(taskParent, updatedTasks);
        }
    }

    return updatedTasks;
};

gantt._updateTaskPosition = function(task, start_date, duration) {
    task.start_date = this.getClosestWorkTime({date: start_date, dir: "future"});
    task.end_date = this.calculateEndDate(start_date, duration);
};

gantt._updateTasks = function(tasks) {
    this._autoscheduling_in_proggress = true;
    gantt.batchUpdate(function() {
        for (var task in tasks) {
            gantt.updateTask(task);
        }
    });
    this._autoscheduling_in_proggress = false;
};

gantt._getTaskMinimalStartDate = function(task) {
    var result = {};
    for (var i = 0; i < task.$target.length; i++) {
        var currentLink = this.getLink(task.$target[i]);
        var currentSource = this.getTask(currentLink.source);
        var links = this.config.links;
        var minStartDateForCurrentTask;
        var realProjDates;

        if(gantt._isProject(task)) {
            realProjDates = gantt.getSubtaskDates(task.id);
            task.start_date = realProjDates.start_date;
            task.duration = gantt.calculateDuration(realProjDates.start_date, realProjDates.end_date);
            task.end_date = realProjDates.end_date;
        }

        if(gantt._isProject(currentSource)) {
            realProjDates = gantt.getSubtaskDates(currentSource.id);
            currentSource.start_date = realProjDates.start_date;
            currentSource.duration = gantt.calculateDuration(realProjDates.start_date, realProjDates.end_date);
            currentSource.end_date = realProjDates.end_date;
        }

        switch (currentLink.type) {
            case links.finish_to_start:
                if (!result.date || result.date < currentSource.end_date) {
                    result.date = currentSource.end_date;
                    result.link = currentLink;
                    result.source_task = currentSource;
                }
                break;

            case links.start_to_start:
                if (!result.date || result.date < currentSource.start_date) {
                    result.date = currentSource.start_date;
                    result.link = currentLink;
                    result.source_task = currentSource;
                }
                break;

            case links.start_to_finish:
                minStartDateForCurrentTask = this.calculateEndDate(currentSource.start_date, (task.duration || 1) * (-1));
                if (!result.date || result.date < minStartDateForCurrentTask) {
                    result.date = minStartDateForCurrentTask;
                    result.link = currentLink;
                    result.source_task = currentSource;
                }
                break;

            case links.finish_to_finish:
                minStartDateForCurrentTask = this.calculateEndDate(currentSource.end_date, task.duration * (-1));
                if (!result.date || result.date < minStartDateForCurrentTask) {
                    result.date = minStartDateForCurrentTask;
                    result.link = currentLink;
                    result.source_task = currentSource;
                }
                break;
            default: break;
        }
    }
    if(task.start_date < result.date){
        result._mustMove = true;
    }
    return result;
};

gantt._getProjectMinimalStartDate = function(project) {
    if(!project) {
        return {};
    }
    var result;

    if(gantt._isProject(project)) {
        result = gantt._getTaskMinimalStartDate(project);
    }
    else {
        result = {};
    }

    var projectParent;
    if(project.parent && project.parent != gantt.config.root_id) {
        projectParent = this.getTask(project.parent);
    }

    var parentMinDate = gantt._getProjectMinimalStartDate(projectParent);
    if(!result.date  || result.date < parentMinDate.date)
    {
        result = parentMinDate;
    }
    return result;
};

gantt._autoScheduleParents = function(parent, updatedTasks, strictStart) {
    if(!parent) {
        return updatedTasks;
    }

    if(parent.parent && parent.parent != gantt.config.root_id) {
        var bigParent = this.getTask(parent.parent);
        gantt._autoScheduleParents(bigParent, updatedTasks, strictStart);
    }
    gantt._autoScheduleLinkedTasks(parent, updatedTasks, 0, strictStart);

    return updatedTasks;
};

gantt._isElementAnStrictStartParent = function(strictStart,elementId)
{
    if(!strictStart) {
        return false;
    }
    var parent;

    while (this.getParent(parent || strictStart))
    {
        parent = this.getParent(parent || strictStart);

        if(parent == elementId) {
            return true;
        }
    }
    return false;
};

gantt._isTask = function (task) {
    return (!task.type || task.type != gantt.config.types.project) && !(task.$no_start || task.$no_end);
};

gantt._isProject = function (task) {
    return !this._isTask(task);
};